// Copyright 2023 The ostui Authors
// SPDX-License-Identifier: GPL-3.0-only

package main

import (
	"errors"
	"fmt"
	"io/fs"
	"log"
	"net/url"
	"os"
	"runtime"
	"runtime/debug"
	"runtime/pprof"
	"sync"
	"time"

	"ser1.net/ostui/logger"
	"ser1.net/ostui/mpvplayer"
	"ser1.net/ostui/remote"
	"ser1.net/ostui/subsonic"

	"ser1.net/clapconf"
	"ser1.net/claptrap/v4"
)

// TODO (D) Update screenshots in the README
// TODO (A) Add mocking library
// TODO (C) Get unit tests up to some non-embarassing percentage
// TODO (C) Support "Download" for songs, albums, artists, and playlists
// TODO (A) Allow fetching credentials from a secret service
// FIXME (B) Adding an album from the browser view where the artist has only one song adds the whole album. Adding the one song manually adds the wrong song. Eilish, No Time to Die

var osExit = os.Exit  // A variable to allow mocking os.Exit in tests
var headlessMode bool // This can be set to true during tests
var testMode bool     // This can be set to true during tests, too

// DEVELOPMENT in case no version can be identified
const DEVELOPMENT = "development"

// APIVersion is the OpenSubsonic API version we talk, communicated to the server
const APIVersion = "1.8.0"

// Name is the client name we tell the server
var Name = "ostui"

// Version is the program version; usually set from BuildInfo
var Version = DEVELOPMENT

// parseConfig takes the first non-flag arguments from flags and parses it
func parseConfig(server *claptrap.CommandConfig, auth *claptrap.CommandConfig, uri string) error {
	if u, e := url.Parse(uri); e == nil {
		// If credentials were provided
		if len(u.User.Username()) > 0 {
			auth.Args["username"].Set(u.User.Username())
		}
		u.User = nil
		server.Args["host"].Set(u.String())
	} else {
		return fmt.Errorf("Invalid server format; must be a valid URL in the format [http[s]://[user:pass@]server:port]")
	}
	return nil
}

// return codes:
// 0 - OK
// 1 - generic errors
// 2 - main config errors
// 2 - keybinding config errors
func main() {
	root := claptrap.Command("ostui", "A terminal client for OpenSubsonic music servers")
	auth := root.AddCommand("auth", "Manually pass in authorization parameters (overrides the config file)")
	auth.Add("--username", "", "OpenSubsonic server user name")
	auth.Add("--password", "", "OpenSubsonic server user password")
	auth.Add("--plaintext", "", "Allow plain text password authentication")
	server := root.AddCommand("server", "Manually pass in server parameters (overrides the config file)")
	server.Add("--host", "127.0.0.1", "OpenSubsonic server host")
	server.Add("--scrobble", "Scrobble listening information")
	client := root.AddCommand("client", "Manually pass in client parameters (overrides the config file)")
	client.Add("--random-songs", "", "Set the number of random songs to select, when asked for")
	uic := root.AddCommand("ui", "Manually pass in ui parameters (overrides the config file)")
	uic.Add("--hide-song-info", "Disable the song info panel in the queue display")
	uic.Add("--spinner", "┤┘┴└├┌┬┐", "Set the spinner characters (see the readme)")
	root.Add("--help", "-h", "Print usage")
	root.Add("--mpris", "Enable MPRIS2")
	root.Add("--list", "list server data")
	root.Add("--playlists", "include playlist info (only used with --list; playlists can take a long time to load)")
	cpuprofile := root.Add("--cpuprofile", "", "write cpu profile to `file`")
	memprofile := root.Add("--memprofile", "", "write memory profile to `file`")
	root.Add("--logfile", "", "Also write log messages to this file")
	root.Add("--config", "", "use config `file`")
	root.Add("--version", "print the ostui version and exit")
	root.Add("hosturi", "The server URI in the format [[user:pass@]server:port] (overrides everything else)")

	root.Parse(nil)
	if root.Bool("help") {
		claptrap.Usage()
		osExit(0)
	}
	if Version == DEVELOPMENT {
		if bi, ok := debug.ReadBuildInfo(); ok {
			Version = bi.Main.Version
		}
	}
	if root.Bool("version") {
		fmt.Printf("ostui %s\n", Version)
		osExit(0)
	}
	if err := clapconf.LoadConfig(root.String("config")); err != nil && errors.Is(err, &fs.PathError{}) {
		fmt.Fprintf(os.Stderr, "Failed to read configuration from file: %v (%T)\n", err, err)
		osExit(2)
	}
	if root.Args["hosturi"].IsSet() {
		// Override everything else with a URI passed on the command line
		if err := parseConfig(server, auth, root.String("hosturi")); err != nil {
			log.Fatal("could not create CPU profile: ", err)
			osExit(1)
		}
	}

	// Claptrap supports required arguments, but clapconf currently does not
	// provide a mechanism for accounting for the assertions when they're not
	// provided on the command line. So, for now we have to check for the
	// parameters manually.
	if auth.String("username") == "" {
		fmt.Printf("username is required, either through the command line or the config file.\n")
		osExit(1)
	}
	if auth.String("password") == "" {
		fmt.Printf("password is required, either through the command line or the config file.\n")
		osExit(1)
	}
	if server.String("host") == "" {
		fmt.Printf("host is required, either through the command line or the config file.\n")
		osExit(1)
	}

	// cpu/memprofile code straight from https://pkg.go.dev/runtime/pprof
	if cpuprofile.IsSet() && cpuprofile.Value.(string) != "" {
		f, err := os.Create(cpuprofile.Value.(string))
		if err != nil {
			log.Fatal("could not create CPU profile: ", err)
		}
		defer f.Close() // error handling omitted for example
		if err := pprof.StartCPUProfile(f); err != nil {
			log.Fatal("could not start CPU profile: ", err)
		}
		defer pprof.StopCPUProfile()
	}

	logger := logger.Init(root.String("logfile"))
	defer logger.Close()

	// init mpv engine
	player, err := mpvplayer.NewPlayer(logger)
	if err != nil {
		fmt.Println("Unable to initialize mpv. Is mpv installed?")
		osExit(1)
	}

	var mprisPlayer *remote.MprisPlayer
	// init mpris2 player control (linux only but fails gracefully on other systems)
	if root.Bool("mpris") {
		mprisPlayer, err = remote.RegisterMprisPlayer(player, logger)
		if err != nil {
			fmt.Printf("Unable to register MPRIS with DBUS: %s\n", err)
			fmt.Println("Try running without MPRIS")
			osExit(1)
		}
		defer mprisPlayer.Close()
	}

	// init macos mediaplayer control
	if runtime.GOOS == "darwin" {
		if err = remote.RegisterMPMediaHandler(player, logger); err != nil {
			fmt.Printf("Unable to initialize MediaPlayer bindings: %s\n", err)
			osExit(1)
		} else {
			logger.Print("MacOS MediaPlayer registered")
		}
	}

	if testMode {
		fmt.Println("Running in test mode for testing.")
		osExit(0x23420001)
		return
	}

	connection := subsonic.Init(logger)
	connection.SetClientInfo(Name, APIVersion)
	connection.Username = auth.String("username")
	connection.Password = auth.String("password")
	connection.Host = server.String("host")
	connection.PlaintextAuth = auth.Bool("plaintext")
	connection.Scrobble = server.Bool("scrobble")
	connection.RandomSongNumber = uint(client.Int("random-songs"))

	artistInd, err := connection.GetArtists()
	if err != nil {
		fmt.Printf("Error fetching indexes from server: %s\n", err)
		osExit(1)
	}
	// Sparse artist information: id, name, albumCount, coverArt, artistImageUrl
	artists := make([]subsonic.Artist, 0)
	artistCount := 0
	albumCount := 0
	for _, ind := range artistInd.Index {
		artistCount += len(ind.Artists)
		for _, art := range ind.Artists {
			albumCount += art.AlbumCount
			artists = append(artists, art)
		}
	}

	if root.Bool("list") {
		var playlists subsonic.Playlists
		var err error
		wg := sync.WaitGroup{}
		wg.Add(1)
		if root.Bool("playlists") {
			go func() {
				playlists, err = connection.GetPlaylists()
				wg.Done()
			}()
		}
		if si, err := connection.GetServerInfo(); err == nil {
			fmt.Printf("Server %-20s: %s\n", "status", si.Status)
			fmt.Printf("Server %-20s: %s\n", "Subsonic API version", si.Version)
			fmt.Printf("Server %-20s: %s\n", "type", si.Type)
			fmt.Printf("Server %-20s: %s\n", "version", si.ServerVersion)
			fmt.Printf("Server %-20s: %t\n", "is OpenSubsonic", si.OpenSubsonic)
		} else {
			fmt.Printf("\n  Error fetching playlists from server: %s\n", err)
		}
		indexes, err := connection.GetIndexes()
		fmt.Printf("%-27s: %d\n", "Indexes", len(indexes.Index))
		directoryCount := 0
		subDirCount := 0
		for _, ind := range indexes.Index {
			directoryCount += len(ind.Artists)
			for _, art := range ind.Artists {
				subDirCount += art.AlbumCount
			}
		}
		fmt.Printf("%-27s: %d\n", "Directories", directoryCount)
		fmt.Printf("%-27s: %d\n", "Subdirectories", subDirCount)
		fmt.Printf("%-27s: %d\n", "Artists", artistCount)
		fmt.Printf("%-27s: %d\n", "Albums", albumCount)
		if st, err := connection.GetStarred(); err == nil {
			fmt.Printf("%-27s: %d\n", "Starred albums", len(st.Albums))
			fmt.Printf("%-27s: %d\n", "Starred artists", len(st.Artists))
			fmt.Printf("%-27s: %d\n", "Starred songs", len(st.Songs))
		}
		if root.Bool("playlists") {
			var spinnerText []rune = []rune(uic.String("spinner"))
			if len(spinnerText) == 0 {
				spinnerText = []rune("┤┘┴└├┌┬┐")
			}
			fmt.Printf("%-27s: %c", "Playlists", spinnerText[0])
			stop := make(chan struct{})
			go func() {
				defer wg.Done()
				spinnerMax := len(spinnerText) - 1
				timer := time.NewTicker(500 * time.Millisecond)
				defer timer.Stop()
				idx := 1
				for {
					select {
					case <-timer.C:
						fmt.Printf("\b%c", spinnerText[idx])
						idx++
						if idx > spinnerMax {
							idx = 0
						}
					case <-stop:
						fmt.Printf("\b")
						return
					}
				}
			}()
			wg.Wait()
			wg.Add(1)
			stop <- struct{}{}
			wg.Wait()
			if err == nil {
				fmt.Printf("%d\n", len(playlists.Playlists))
				for _, pl := range playlists.Playlists {
					if len(pl.Entries) > 0 {
						fmt.Printf("  %25s: %d\n", pl.Name, len(pl.Entries))
					}
				}
			} else {
				fmt.Printf("\n  Error fetching playlists from server: %s\n", err)
			}
		}

		osExit(0)
	}

	if headlessMode {
		fmt.Println("Running in headless mode for testing.")
		osExit(0)
		return
	}

	ui := InitGui(artists, connection, player, logger, mprisPlayer)

	// run main loop
	if err := ui.Run(); err != nil {
		// FIXME print a useful message and exit, don't panic
		panic(err)
	}

	if memprofile.IsSet() && memprofile.Value.(string) != "" {
		f, err := os.Create(memprofile.Value.(string))
		if err != nil {
			log.Fatal("could not create memory profile: ", err)
		}
		defer f.Close() // error handling omitted for example
		runtime.GC()    // get up-to-date statistics
		if err := pprof.WriteHeapProfile(f); err != nil {
			log.Fatal("could not write memory profile: ", err)
		}
	}
}
